/**
  Physics simulator demo program.
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2011-01-20 (Public Domain)
*/
#include "physics/world.h"
#include "ogl/glsl.h"
#include "soil/SOIL.h"

/* LAME! just include library bodies here, for easy linking */
#include "physics/world.cpp" 
#include "physics/config.cpp"
#include "ogl/glsl.cpp"
#include "soil/SOIL.c"
#include "soil/stb_image_aug.c"

const float far_bumper=108.0; // units to far red bumper

/* One simulated particle */
class particle {
public:
	vec4 color; // RGBA
	vec3 pos, vel; // in world coordinates-- meters
	vec3 netforce; // in newtons (quarter-pounds)
	float mass; // in kilograms
	
	void start_step(double dt) {
		/*
		 For stability, I'm using "Leapfrog" integration:
			http://en.wikipedia.org/wiki/Leapfrog_integration
		 Basically, we just update position, *then* calculate
		 net force before moving the velocity.
		*/
		pos+=dt*vel;
		
		/* Zero out net force, so springs can add to it */
		netforce=vec3(0.0);
	}
	
	void finish_step(double dt,const vec3 &target) {
	/* Newtonian motion */
		vec3 acc=netforce/mass; 

		acc.z+=-9.8;  /* gravity */
		
		/* User interface force */
		vec3 fear=-(target-pos);
		if (length(fear)<1.0) { // close enough to scare me! 
			acc+=normalize(fear)*1000.0;
		}
		
		/* "wind drag": fairly gentle */
		acc+=-0.2*vel;
		vel+=dt*acc;

		if (pos.z<0.0) { /* hit the ground */
			vel.z=0.6*fabs(vel.z);
			vel*=0.4; // ground drag
			pos.z=0.0;
			//printf("g");
		}
		if (pos.y>far_bumper) { /* hit the far wall */
			vel.y=-0.6*fabs(vel.y);
			pos.y=far_bumper;
		}
	}
};


vec3 make2d(vec3 src) {
	src.z=0.0;
	return src;
}

const float wheelRadius=0.4, wheelThickness=0.3; /* in meters */

/* An axle with attached wheel */
class axleAndWheel {
public:
	int outside,inside; /* axle's vertex numbers (in p) */
	std::vector<particle> &p;
	
	/* Wheel's coordinate system: 
		origin is center of hub (p[outside].pos)
		Z points down the axle
		Y points up (against gravity)
		X points forward/backward
	*/
	vec3 wo,wx,wy,wz;
	vec3 stable_friction; /* averaged over a few timesteps, for stability */
	
	/* Wheel's amount of rotation, in radians */
	double orient;
	
	/* Wheel's steering angle, in radians */
	double steer;
	
	/* Turning torque applied to wheel, in kg-m */
	double torque;
	
	axleAndWheel(int outside_,int inside_,std::vector<particle> &p_)
		:outside(outside_), inside(inside_), p(p_) 
	{
		orient=torque=steer=0.0;
		stable_friction=vec3(0.0);
	}
	
	
	void simulate(double dt,physics::library &lib) {
		/* Ground height interaction */
		float touching=0.0;
		if (p[outside].pos.z<=wheelRadius*1.5) {
			touching=1.0; /* <- for stability, give it a little wiggle room */
			if (p[outside].pos.z<=wheelRadius) {
				p[outside].pos.z=wheelRadius; /* push up (hard constraint) */
				p[outside].vel.z*=0.3; /* suck up collision energy */
			}
		}
	
		/* Coordinate system */
		wo=p[outside].pos;
		vec3 axle=wz=normalize(p[inside].pos-wo);
		wy=vec3(0,0,1);
		wx=normalize(cross(wy,wz)+steer*wz);
		wy=normalize(cross(wz,wx));
		wz=normalize(cross(wx,wy));
		
		/* Compute velocity versus ground (assumed to be stationary) */
		vec3 relvel=p[outside].vel; 
		
		/* roll the wheel */
		orient-=dt*(torque*10.0+touching*dot(relvel,wx)/wheelRadius);
		
		/* Apply friction forces with ground */
		float coefficientoffriction=1000.0; /* newtons force per m/s velocity */
		
		vec3 friction=dot(relvel,wz)*wz; // ground friction (force along axle)
		stable_friction=0.9*stable_friction+0.1*friction;
		
		/* scale back motor output at high RPM: 
		  too fast, and you tend to flip */
		const float peakvel=22.0; /* meters/second */
		float peakscl=1.0-torque*dot(relvel,wx)/peakvel; 
		if (peakscl>1.0) peakscl=1.0; /* otherwise rev-to-fwd pops a wheelie! */
		
		vec3 netforce=coefficientoffriction*touching*
			(
				//-0.1*relvel // axle drag (more stability, like molasses!)
				-0.4*stable_friction // ground friction (force along axle)
				+1.8*torque*peakscl*wx // motor torque (peak vel, 'wheelie')
			);	
		p[outside].netforce+=netforce;
	}
	
	void draw(physics::library &lib) {
		glPushMatrix();
		/* Load up tire texture with SOIL */
		static GLuint texture=SOIL_load_OGL_texture(
			"tire.jpg",0,0,SOIL_FLAG_MIPMAPS
		);
		glBindTexture(GL_TEXTURE_2D,texture);

		/* Draw our coordinate axes for debugging */
		glLineWidth(20.0);
		glBegin(GL_LINES);
		//glColor3f(1,0,0); glVertex3fv(wo); glVertex3fv(wo+wx);
		//glColor3f(0,1,0); glVertex3fv(wo); glVertex3fv(wo+wy);
		glColor3f(0,0,1); glVertex3fv(wo); glVertex3fv(wo+0.3*wz);
		glEnd();

		/* Draw tire, then shadow */
		for (int shadow=0;shadow<=1;shadow++) {
		  if (shadow) {
		  	glDisable(GL_TEXTURE_2D);
			glDepthMask(GL_FALSE);
			glColor4f(0.0f,0.0f,0.0f,0.3f);
			glScalef(1.0f,1.0f,0.0f); /* squish shadows down */
		  } else { /* normal tire */
			glEnable(GL_TEXTURE_2D);
			glColor3f(1,1,1);
		  }
		  
		  /* Draw little pie slices around tire */
		  glBegin(GL_TRIANGLE_STRIP);
		  double da=1.0/16; /* size of a tire slice, in revolutions */
		  for (double angle=0.0;angle<1.0;angle+=da) {
			double a1=angle*2.0*M_PI;
			double a2=(angle+da)*2.0*M_PI;
			double s1=sin(a1+orient), c1=cos(a1+orient); /* coords rotate with tire */
			double s2=sin(a2+orient), c2=cos(a2+orient);
			double ts1=sin(a1), tc1=cos(a1); /* texcoords don't rotate */
			double ts2=sin(a2), tc2=cos(a2);
			glTexCoord2f(0.5f,0.5f);
			glVertex3fv(wo);
			glVertex3fv(wo); /* doubling start, to reset tri strip */
			
			glTexCoord2f(tc1*0.45+0.5,ts1*0.45+0.5);
			glVertex3fv(wo+wheelRadius*(c1*wx+s1*wy));
			glTexCoord2f(tc2*0.45+0.5,ts2*0.45+0.5);
			glVertex3fv(wo+wheelRadius*(c2*wx+s2*wy));
			
			vec3 out=-wz*wheelThickness;
			
			glTexCoord2f(tc1*0.5+0.5,ts1*0.5+0.5);
			glVertex3fv(wo+wheelRadius*(c1*wx+s1*wy)+out);
			glTexCoord2f(tc2*0.5+0.5,ts2*0.5+0.5);
			glVertex3fv(wo+wheelRadius*(c2*wx+s2*wy)+out);
			
			glTexCoord2f(0.5f,0.5f);
			glVertex3fv(wo+out);
			glVertex3fv(wo+out); /* doubling end, to reset */
			
		  }
		  glEnd();
		}
		glDisable(GL_TEXTURE_2D);
		glDepthMask(GL_TRUE);
		glPopMatrix();
	}
};


/* Random float between 0 and 1 */
float randf(void) {
	float f=rand();
	return f/RAND_MAX;
}

/* Random vector where x y and z vary between 0 and 1 */
vec3 randVec3(void) {
	return vec3(randf(),randf(),randf());
}

/* A 3D tetrahedron */
class tet {
public:
	int v[4]; /* each int is a vertex index */
	float restlength[6]; // length in meters of rest position of each edge
};

void bad(const char *err) {
	std::cout<<"Fatal file read error:  "<<err<<"\n";
	exit(1);
}

/* A particle system */
class particlesystem : public physics::object {
	std::vector<particle> p;
	std::vector<tet> tets;
	vec3 targetpoint;
	bool initRestState;
	int frontVertex, backVertex;
	
	std::vector<axleAndWheel *> axles;
	
	/* Build axle objects based on these vertex numbers */
	void makeAxles(const int *vAxle,int nAxleVtx) {
		int verts[4][2];
		int vertcount[4];
		for (int quadrant=0;quadrant<4;quadrant++) vertcount[quadrant]=0;
		for (int v=0;v<nAxleVtx;v++) {
			/* Figure out which quadrant we're in */
			int x=(p[vAxle[v]].pos.x<0.0)?0:1;
			int y=(p[vAxle[v]].pos.y<0.0)?0:1;
			int quadrant=x+2*y; /* 0 through 3 */
			verts[quadrant][vertcount[quadrant]++]=vAxle[v];
			if (vertcount[quadrant]>1) bad("Can't locate all 4 axles!\n");
		}
		for (int quadrant=0;quadrant<4;quadrant++) {
			if (vertcount[quadrant]!=1) bad("Can't locate all 4 axles!\n");
			int a=verts[quadrant][0]; /* outside vertex */
			int b=verts[quadrant^1][0]; /* inside vertex */
			//if (fabs(p[a].pos.x)<fabs(p[b].pos.x)) std::swap(a,b);
			axles.push_back(new axleAndWheel(a,b,p));
		}
		
	}
public:
	particlesystem(void) 
		:physics::object(0.001)  /* <- our timestep, in seconds */
	{
		initRestState=true;
		
		/* Read a TetGen set of tetrahedra
		      http://tetgen.berlios.de/
		*/
		std::ifstream f("model.1.ele");
		int ntet,npertet,nedge;
		f>>ntet>>npertet>>nedge;
		if (npertet!=4) bad("expected 4 vertices per tet");
		
		/* Read all the tets */
		for (int i=0;i<ntet;i++) {
			tet t;
			int tetno; f>>tetno; if (i!=tetno) bad("Unexpected tet number!");
			for (int j=0;j<npertet;j++)
				f>>t.v[j];
			tets.push_back(t);
		}
		if (!f) bad("I/O error reading .ele file");
		f.close();
		
		
		/* Silly hack: locate axle vertices based on positions */
		int nAxle=0;
		const int maxAxle=8;
		int vAxle[maxAxle]; /* vertex numbers of each axle */
		
		/* Read all the XYZ vertices */
		f.open("model.1.node");
		int nnode, coordpernode, val1pernode, val2pernode;
		f>>nnode>> coordpernode>> val1pernode>> val2pernode;
		vec3 minpos=vec3(+1.0e10,+1.0e10,+1.0e10);
		vec3 maxpos=vec3(-1.0e10,-1.0e10,-1.0e10);
		for (int i=0;i<nnode;i++) {
			int nodeno; f>>nodeno; if (nodeno!=i) bad("Unexpected node number!");
			float x,y,z;
			f>>x>>y>>z;
			particle np;
			np.color=1.0*randVec3(); np.color.w=0.5; /* alpha */
			np.pos=vec3(x,-z,-y);
			if (np.pos.z<-1.11) {/* that's an axle vertex */
				if (nAxle<maxAxle) vAxle[nAxle]=i;
				nAxle++; 
			}
			minpos=min(np.pos,minpos);
			maxpos=max(np.pos,maxpos);
			
			np.vel=vec3(-2.0,-2.0,2.0); // (randVec3()-vec3(0.5));
			p.push_back(np);
		}
		if (!f) bad("I/O error reading .node file");
		f.close();
		
		/* Figure out which axle is which */
		printf("Found %d axle vertices (z range %.3f - %.3f)\n",nAxle,minpos.z,maxpos.z);
		makeAxles(vAxle,nAxle);
		
		/* Now that we've located the axles, push the object to the starting point */
		for (int i=0;i<nnode;i++) {
			if (p[i].pos.y==maxpos.y) frontVertex=i; /* <- used for follow camera */
			if (p[i].pos.y==minpos.y) backVertex=i;
			
			/* make underbody heavier (to reduce rollovers) */
			p[i].mass=0.5+2.0*(maxpos.y-p[i].pos.y)/(maxpos.y-minpos.y);
		
			/* push up off the ground initially */
			p[i].pos+=vec3(3.0,0.0,1.4+wheelRadius);
		}
	}
	
	void simulate(physics::library &lib) {
		float slowmo=1.0;
		float dt=timestep.dt*slowmo;
	
		// Start timesteps
		for (unsigned int i=0;i<p.size();i++)
		{
			p[i].start_step(dt);
			if (length(p[i].vel)>1000.0) {std::cout<<"Unstable at t="<<timestep.last_t<<"\n"; exit(1);}
		}
		
		// Run physics on axles
		for (unsigned int i=0;i<axles.size();i++)
			axles[i]->simulate(dt,lib);
		
		// Loop over edges, add in net forces to particles
		for (unsigned int i=0;i<tets.size();i++) {
			int edgeno=0;
			for (int j=0;j<3;j++) 
			for (int k=j+1;k<4;k++, edgeno++) {
				int a=tets[i].v[j];
				int b=tets[i].v[k];
				
				vec3 dist=p[a].pos-p[b].pos;
				vec3 along=normalize(dist);
				float len=length(dist);
				
				/* Newtons of force per meter of displacement:
				    this should be pretty huge: in the thousands
				*/
				float springconst=30000.0;
				
				if (initRestState) tets[i].restlength[edgeno]=len-0.0001;
				
				float restlen=tets[i].restlength[edgeno];
			#if 1 /* breakable springs */
				if (restlen<0.0) continue; /* we are already broken */
				
				if (fabs(len-restlen)>0.10) { /* we are now broken! */
					tets[i].restlength[edgeno]=-1.0; 
				}
			#endif
				
				/* Spring force is straight from Hooke's Law */
				float f=-springconst*(len-restlen);
				
				p[a].netforce+=f*along;
				p[b].netforce-=f*along;
			}
		}
		initRestState=false;
		
		// Finish timestep
		for (unsigned int i=0;i<p.size();i++) {
			p[i].finish_step(dt,targetpoint);
		}
	}
	
	void draw(physics::library &lib) {
		if (!lib.key_toggle['f']) { /* follow camera */
			targetpoint=vec3(9999.9); /* disable pusher sphere */
			
			/* Lock camera to be behind model */
			vec3 towardcamera=p[backVertex].pos-p[frontVertex].pos;
			camera=p[backVertex].pos +0.2*(towardcamera) +vec3(0,0,4.0);
			if (dot(towardcamera,vec3(0,0,1))>-0.5)
				camera_orient.x=normalize(cross(vec3(0,0,1),towardcamera));
			
			/* zero all axle torques (should this be in simulate above?) */
			for (unsigned int i=0;i<axles.size();i++) {
				axles[i]->torque=0.0;
				axles[i]->steer=0.0;
			}
			
			float t=1.0, forward=0.0; /* torque, for driving */
			if (lib.key['w']) { forward+=t; }
			if (lib.key['s']) { forward-=t; }
			axles[0]->torque+=forward; axles[1]->torque-=forward;
			
			float s=0.3, turn=0.0; /* steering */
			if (lib.key['a']) { turn-=s; } /* turn left */
			if (lib.key['d']) { turn+=s; } /* turn right */
			//axles[0]->steer-=turn; axles[1]->steer-=turn; // 4 wheel steering
			axles[2]->steer+=turn; axles[3]->steer+=turn;
			
		} 
		else { /* else 'f' key pressed: free camera and pusher sphere */
			targetpoint=camera-camera_orient.z*1.0-camera_orient.y*1.0;
		}
		
		// Draw axles
		for (unsigned int i=0;i<axles.size();i++)
			axles[i]->draw(lib);
	
		/* Draw all the tets as quads (a HACK, but fairly pretty!)*/
		for (int shadow=1;shadow>=0;shadow--) {
			glPushMatrix();
			if (shadow) { /* shadow only */
				glDepthMask(GL_FALSE);
				glColor4f(0.0f,0.0f,0.0f,0.05f);
				glScalef(1.0f,1.0f,0.0f);
			}
			else { /* normal drawing */
				glDepthMask(GL_TRUE);
			}
			glBegin(GL_QUADS);
			for (unsigned int i=0;i<tets.size();i++) {
				for (int j=0;j<4;j++) {
					if (!shadow) glColor4fv(p[tets[i].v[j]].color);
					glVertex3fv(p[tets[i].v[j]].pos);
				}
			}
			glEnd();
			glPopMatrix();
		}
		
		/* Green sphere is our pusher */
		glLineWidth(2.0);
		for (int shadow=0;shadow<=1;shadow++) {
			glPushMatrix();
			glColor3f(0,1,0);
			if (shadow) {
				glScalef(1,1,0);
				glColor4f(0,0,0,0.3);
			}
			glTranslatef(VEC3_TO_XYZ(targetpoint));
			glutWireSphere(1.0,6,8);
			glPopMatrix();
		}
		
		/* Red quad represents the bumper */
		glLineWidth(1.0);
		for (int outline=0;outline<=1;outline++) {
			glPushMatrix();
			glTranslatef(0.0,far_bumper,+1);
			if (outline) glDepthMask(GL_TRUE); else glDepthMask(GL_FALSE);
			if (outline) glColor3f(0,0,0); else glColor4f(1,0,0,0.3);
			if (outline) glBegin(GL_LINE_LOOP); else glBegin(GL_QUADS);
				glVertex3f(-500,0,-4);glVertex3f(+500,0,-4);
				glVertex3f(+500,0,+4);glVertex3f(-500,0,+4);
			glEnd();
			glPopMatrix();
		}
	}
};


/* Called create a new simulation */
void physics_setup(physics::library &lib) {
	if (!lib.key_toggle['f']) {
		//camera=vec3(0.0,-5.0,1.7);
		camera_orient.x=vec3(1,0,0);
		camera_orient.z=normalize(vec3(0,-0.6,0.5));
	}
	lib.world->add(new particlesystem());
	lib.background[0]=lib.background[1]=0.6;
	lib.background[2]=1.0; // light blue
}

/* Called every frame to draw everything */
void physics_draw_frame(physics::library &lib) {
	/* use SOIL ground texture */
	static GLuint texture=SOIL_load_OGL_texture(
		"ground.jpg",0,0,SOIL_FLAG_MIPMAPS|SOIL_FLAG_TEXTURE_REPEATS
	);
	glBindTexture(GL_TEXTURE_2D,texture);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAX_ANISOTROPY_EXT,4);
	/* Draw opaque white ground plane */
	int groundsize=500;
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_DEPTH_TEST);
	glColor3f(1,1,1);
	glPushMatrix();glTranslatef(0,0,-0.01); /* <- shove down, to reduce Z fighting */
	glBegin(GL_QUADS);
	glTexCoord2i(-groundsize,-groundsize); glVertex2i(-groundsize,-groundsize);
	glTexCoord2i(+groundsize,-groundsize); glVertex2i(+groundsize,-groundsize);
	glTexCoord2i(+groundsize,+groundsize); glVertex2i(+groundsize,+groundsize);
	glTexCoord2i(-groundsize,+groundsize); glVertex2i(-groundsize,+groundsize);
	glEnd();
	glPopMatrix();
	glDisable(GL_TEXTURE_2D);

	/* Point and line smoothing needs normal alpha blending */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	/* Now draw objects */
	lib.world->draw(lib);
}


